/*
 * serio.c - Serial device routines.
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/serial.h>

#include "macros.h"
#include "log.h"
#include "xmalloc.h"
#include "list.h"
#include "serio.h"

struct serio_info {
	int		fd;
	struct termios	termio;
	struct termios	termio_old;
	int		is_nonblock: 1;
	int		is_canon: 1;
};

static slist_t serio_list;

static struct serio_info *serio_find(int fd)
{
	slist_t list;

	for (list = serio_list; list != NULL; list = list->next) {
		struct serio_info *si = (struct serio_info *) list->data;
		if (si->fd == fd) {
			return si;
		}
	}
	return NULL;
}

static struct serio_info *serio_remove(int fd)
{
	slist_t list;

	for (list = serio_list; list != NULL; list = list->next) {
		struct serio_info *si= (struct serio_info *) list->data;
		if (si->fd == fd) {
			slist_unlink(serio_list, list);
			/* destroy container */
			slist_free(&list);
			return si;
		}
	}

	return NULL;
}

speed_t serio_str2speed(const char *str)
{
	if (!str)
		error("serio_str2speed: NULL str");

	if (strcmp(str, "50") == 0)
		return B50;
	else if (strcmp(str, "75") == 0)
		return B75;
	else if (strcmp(str, "110") == 0)
		return B110;
	else if (strcmp(str, "134") == 0)
		return B134;
	else if (strcmp(str, "150") == 0)
		return B150;
	else if (strcmp(str, "200") == 0)
		return B200;
	else if (strcmp(str, "300") == 0)
		return B300;
	else if (strcmp(str, "600") == 0)
		return B600;
	else if (strcmp(str, "1200") == 0)
		return B1200;
	else if (strcmp(str, "1800") == 0)
		return B1800;
	else if (strcmp(str, "2400") == 0)
		return B2400;
	else if (strcmp(str, "4800") == 0)
		return B4800;
	else if (strcmp(str, "9600") == 0)
		return B9600;
	else if (strcmp(str, "19200") == 0)
		return B19200;
	else if (strcmp(str, "38400") == 0)
		return B38400;
	else if (strcmp(str, "57600") == 0)
		return B57600;
#ifdef B115200
	else if (strcmp(str, "115200") == 0)
		return B115200;
#endif
#ifdef B230400
	else if (strcmp(str, "230400") == 0)
		return B230400;
#endif
	return B115200;
}

int serio_open(const char *file, speed_t speed, serio_csize_t csize, serio_parity_t parity, serio_stopb_t stopbits, serio_flow_t flow)
{
	int fd;
	int res, flags;
	struct termios old, now;
	struct serio_info *si;

	if (file == NULL)
		return -1;

	fd = open(file, O_RDWR | O_NOCTTY | O_NONBLOCK);

	if (fd >= 0) {
		flags = fcntl(fd, F_GETFD);
		if (flags >= 0) {
			res = fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
			if (res == -1) {
				warn("serio_open: fcntl() fd=%d F_SETFD: %s", fd, strerror(errno));
			}
		} else {
			warn("serio_open: fcntl() fd=%d F_GETFD: %s", fd, strerror(errno));
		}
	} else {
		debug("serio_open: open() '%s' res=%d: %s", file, fd, strerror(errno));
		return -1;
	}

	res = tcgetattr(fd, &now);

	if (res == -1) {
		debug("serio_open: tcgetattr() fd=%d: %s", fd, strerror(errno));
		do {
			res = close(fd);
		} while (res < 0 && errno == EINTR);
		return -1;
	}

	old = now;
	now.c_cflag |= CLOCAL;
	cfsetospeed(&now, speed);
	cfsetispeed(&now, speed);

	if (flow == SERIO_FLOW_SOFTWARE) {
		now.c_cflag &= ~CRTSCTS;
		now.c_iflag |= IXON | IXOFF;
	} else if (flow == SERIO_FLOW_HARDWARE) {
		now.c_cflag |= CRTSCTS;
		now.c_iflag &= ~(IXON | IXOFF);
	} else {
		now.c_cflag &= ~CRTSCTS;
		now.c_iflag &= ~(IXON | IXOFF);
	}

	if (csize == SERIO_CSIZE_5) {
		now.c_cflag &= ~(CS8 | CS7 | CS6);
		now.c_cflag |= CS5;
	} else if (csize == SERIO_CSIZE_6) {
		now.c_cflag &= ~(CS8 | CS6 | CS5);
		now.c_cflag |= CS6;
	} else if (csize == SERIO_CSIZE_7) {
		now.c_cflag &= ~(CS8 | CS6 | CS7);
		now.c_cflag |= CS7;
	} else {
		now.c_cflag &= ~(CS7 | CS6 | CS5);
		now.c_cflag |= CS8;
	}

	if (parity == SERIO_PARITY_EVEN) {
		now.c_cflag |= PARENB;
		now.c_cflag &= ~PARODD;
		now.c_iflag &= ~(INPCK | PARMRK);
	} else if (parity == SERIO_PARITY_ODD) {
		now.c_cflag |= (PARENB | PARODD);
		now.c_iflag &= ~(INPCK | PARMRK);
	} else {
		now.c_cflag &= ~(PARENB | PARODD);
		now.c_iflag &= ~(INPCK | PARMRK);
		now.c_iflag |= IGNPAR;
	}

	if (stopbits == SERIO_TWO_STOPB) {
		now.c_cflag |= CSTOPB;
	} else {
		now.c_cflag &= ~CSTOPB;
	}

	now.c_iflag &= ~(IGNBRK | BRKINT | ISTRIP | INLCR | IGNCR | ICRNL);
	now.c_oflag &= ~OPOST;
	now.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	now.c_cc[VMIN] = 1;
	now.c_cc[VTIME] = 0;

	res = tcsetattr(fd, TCSANOW, &now);

	if (res == -1) {
		debug("serio_open: tcsetattr() fd=%d: %s", fd, strerror(errno));
		do {
			res = close(fd);
		} while (res < 0 && errno == EINTR);
		return -1;
	}

	si = xmalloc0(sizeof(*si));
	si->is_canon = FALSE;
	si->is_nonblock = TRUE;
	si->fd = fd;
	si->termio = now;
	si->termio_old = old;
	slist_t list = slist_new();
	list->data = si;
	serio_list = slist_add(serio_list, list);

	debug("serio_open: '%s' fd=%d OK", file, fd);

	return fd;
}

void serio_enable_lowlatency(int fd)
{
	int res;
	struct serio_info *si;
	struct serial_struct ss;

	if (fd < 0)
		error("serio_enable_lowlatency: fd=%d??", fd);

	si = serio_find(fd);

	if (si) {
		res = ioctl(fd, TIOCGSERIAL, &ss);
		if (res) {
			warn("ioctl() TIOCGSERIAL fd=%d: %s", fd, strerror(errno));
			return;
		}
		ss.flags |= ASYNC_LOW_LATENCY;
		res = ioctl(fd, TIOCSSERIAL, &ss);
		if (res) {
			warn("ioctl() TIOCSSERIAL fd=%d: %s", fd, strerror(errno));
			return;
		}
	} else {
		warn("serio_enable_lowlatency: fd=%d not found", fd);
	}
}

int serio_close(int fd)
{
	int res;
	struct serio_info *si;

	if (fd < 0)
		error("serio_close: fd=%d??",  fd);

	si = serio_remove(fd);

	if (si != NULL) {
		res = tcsetattr(fd, TCSANOW, &si->termio_old);
		if (res == -1) {
			warn("serio_close: tcsetattr() fd=%d: %s", si->fd, strerror(errno));
		}
		do {
			res = close(fd);
		} while (res < 0 && errno == EINTR);
	} else {
		warn("serio_close: fd=%d not found", fd);
		return -1;
	}

	/* Free this memory we don't need it any more. */
	xfree(si);

	return 0;
}

void serio_set_canon(int fd)
{
	int res;
	struct termios tio;
	struct serio_info *si;

	if (fd < 0)
		error("serio_set_canon: fd=%d??", fd);

	si = serio_find(fd);

	if (si != NULL) {
		if (!si->is_canon) {
			tio = si->termio;
			tio.c_oflag |= OPOST;
			tio.c_lflag |= (ECHO | ICANON | IEXTEN);
			res = tcsetattr(fd, TCSANOW, &tio);
			if (res == 0) {
				si->termio = tio;
				si->is_canon = TRUE;
			} else {
				warn("serio_set_canon: tcsetattr() fd=%d: %s", fd, strerror(errno));
			}
		}
	} else {
		warn("serio_set_canon: fd=%d not found", fd);
	}
}

void serio_set_raw(int fd)
{
	int res;
	struct termios tio;
	struct serio_info *si;

	if (fd < 0)
		error("serio_set_raw: fd=%d??", fd);

	si = serio_find(fd);

	if (si != NULL) {
		if (si->is_canon) {
			tio = si->termio;
			tio.c_oflag &= ~OPOST;
			tio.c_lflag &= ~(ECHO | ICANON | IEXTEN);
			res = tcsetattr(fd, TCSANOW, &tio);
			if (res == 0) {
				si->termio = tio;
				si->is_canon = FALSE;
			} else {
				warn("serio_set_raw: tcsetattr() fd=%d: %s", fd, strerror(errno));
			}
		}
	} else {
		warn("serio_set_raw: fd=%d not found", fd);
	}
}

void serio_fd_nonblock(int fd)
{
	int res;
	int flags;
	struct serio_info *si;

	if (fd < 0)
		error("serio_fd_nonblock: fd=%d??", fd);

	si = serio_find(fd);

	if (si != NULL) {
		if (!si->is_nonblock) {
			flags = fcntl(fd, F_GETFL);
			if (flags >= 0) {
				flags |= O_NONBLOCK;
				res = fcntl(fd, F_SETFL, flags);
				if (res == 0)
					si->is_nonblock = TRUE;
				else
					warn("serio_fd_nonblock: fcntl() F_SETFL fd=%d: %s", fd, strerror(errno) );
			} else
				warn("serio_fd_nonblock: fcntl() F_GETFL fd=%d: %s", fd, strerror(errno));
		}
	} else {
		warn("serio_fd_nonblock: fd=%d not found", fd);
	}
}

void serio_fd_block(int fd)
{
	int res;
	int flags;
	struct serio_info *si;

	if (fd < 0)
		error("serio_fd_block: fd=%d??", fd);

	si = serio_find(fd);

	if (si != NULL) {
		if (si->is_nonblock) {
			flags = fcntl(fd, F_GETFL);
			if (flags >= 0) {
				flags &= ~O_NONBLOCK;
				res = fcntl(fd, F_SETFL, flags);
				if (res == 0)
					si->is_nonblock = FALSE;
				else
					warn("serio_fd_block: fcntl() F_SETFL fd=%d: %s", fd, strerror(errno));
			} else
				warn("serio_fd_block: fcntl() F_GETFL fd=%d: %s", fd, strerror(errno));
		}
	} else {
		warn("serio_fd_block: fd=%d not found", fd);
	}
}
